/**/

#include "alterator_manager_backends.h"

static GHashTable *backends_data = NULL;

/* Validates a valid name for a d-bus method or parameter. The name can only
contain letters, numbers, and underscores.
It is assumed that the length of the name cannot be greater than 100 and less
than 2. */
static gboolean is_correct_name(gchar* name) {
    if (name == NULL) {
        return FALSE;
    }

    if (g_utf8_strlen(name, 3) < 2) {
        return FALSE;
    }

    for (int i = 0; name[i] != '\0'; i++) {
        if (!g_ascii_isalnum(name[i]) && name[i] != '_') {
            return FALSE;
        }

        if (i >= MAX_NAME_LENGTH - 1) {
            return FALSE;
        }
    }

    return TRUE;
}

static GPtrArray *read_backend_file_names(const gchar *key_files_path) {
    GDir *dir = NULL;
    GError *error = NULL;
    GPtrArray *file_names = NULL;
    gboolean is_correct_suffix = FALSE;

    file_names = g_ptr_array_new_full(5, g_free);
    if (file_names == NULL) {
        g_warning("read_backend_file_names(), g_ptr_array_new() "
                  "returned NULL.");
        return NULL;
    }

    dir = g_dir_open(key_files_path, 0, &error);

    if (error) {
        g_warning("Open directory %s, %s.", key_files_path, error->message);
        g_error_free(error);
        g_ptr_array_free(file_names, TRUE);
        return NULL;
    }

    const gchar *file_name = NULL;
    while ((file_name = g_dir_read_name(dir)) != NULL) {

        is_correct_suffix = g_str_has_suffix(file_name, KEY_FILE_NAME_SUFFIX);

        if (is_correct_suffix) {
            g_debug("Found backend file: %s%s.", key_files_path, file_name);
            g_ptr_array_add(file_names, g_strdup(file_name));
        }
    }

    g_dir_close(dir);
    return file_names;
}

static void interface_info_free(void *interface_info) {
    InterfaceObjectInfo *info = (InterfaceObjectInfo *)interface_info;

    if (info == NULL) {
        return;
    }

    g_free(info->module_name);
    //GHashTable *interface_introspection; //belongs to the module
    if (info->methods != NULL) {
        g_hash_table_unref(info->methods);
    }
    //GDBusInterfaceVTable *interface_vtable; //belongs to the module
    g_free(info);
}

static void hashtable_free(void *hashtable) {
    GHashTable *table = (GHashTable *) hashtable;

    g_hash_table_destroy(table);
}

/* Validates a valid name for an action_id name. The name can only contain
   letters, numbers, and dashes. It is assumed that the length of the name
   cannot be greater than 100 and less than 1. */
static gboolean is_correct_action_id(gchar* name) {
    if (name == NULL) {
        return FALSE;
    }

    if (g_utf8_strlen(name, 2) < 1) {
        return FALSE;
    }

    for (int i = 0; name[i] != '\0'; i++) {
        if (!g_ascii_isalnum(name[i]) && name[i] != '-' && name[i] != '.') {
            return FALSE;
        }

        if (i >= MAX_NAME_LENGTH - 1) {
            return FALSE;
        }
    }

    return TRUE;
}

static void save_info_from_backend_file(GKeyFile *backend_file,
                                        gchar *file_name)
{
    GHashTable *interfaces = NULL;
    InterfaceObjectInfo *interface_info = NULL;
    gchar *type_name;
    gchar *module_name;
    gchar *node_name;
    gchar *interface_name;
    gint thread_limit;

    if (backend_file == NULL) {
        return;
    }

    if (!g_key_file_has_group(backend_file, TOP_SECTION)) {
        g_warning("backend_file (%s) doesn't have group %s.",
                  file_name, TOP_SECTION);
        return;
    }

    type_name = g_key_file_get_string(backend_file,
                                      TOP_SECTION,
                                      TYPE_NAME,
                                      NULL);

    if (type_name == NULL) {
        g_warning("backend_file (%s) has Type == NULL in %s group.",
                  file_name, TOP_SECTION);
        return;
    }

    if (g_strcmp0(type_name, TYPE_BACKEND)) {
        g_free(type_name);
        return;
    }
    g_free(type_name);

    gsize length = 0;
    gchar **groups = g_key_file_get_groups(backend_file, &length);

    if (length < 2) {
        g_warning("backend_file (%s) has less than two groups including the "
                  "%s.", file_name, TOP_SECTION);
        g_strfreev(groups);
        return;
    }

    module_name = g_key_file_get_string(backend_file,
                                        TOP_SECTION,
                                        MODULE_NAME,
                                        NULL);

    if (module_name == NULL) {
        g_warning("backend_file (%s) has module_name == NULL in %s group.",
                  file_name, TOP_SECTION);
        g_strfreev(groups);
        return;
    }

    thread_limit = g_key_file_get_integer(backend_file,
                                         TOP_SECTION,
                                         THREAD_LIMIT_NAME,
                                         NULL);

    if (thread_limit <= 0) {
        g_debug("backend_file (%s) has thread_limit <= 0 in %s group. "
                "This field may not have been set or may have been set "
                "incorrectly. It will be given a default value (%d).",
                file_name, TOP_SECTION, THREAD_LIMIT);
        thread_limit = THREAD_LIMIT;
    }

    node_name = g_key_file_get_string(backend_file,
                                      TOP_SECTION,
                                      NODE_NAME,
                                      NULL);

    if (node_name == NULL) {
        g_warning("backend_file (%s) has node_name == NULL in %s group.",
                  file_name, TOP_SECTION);
        g_strfreev(groups);
        g_free(module_name);
        return;
    }

    if (!is_correct_name(node_name)) {
        g_warning("Incorrect node name (%s) in backend file (%s).", node_name,
                  file_name);
        g_strfreev(groups);
        g_free(module_name);
        g_free(node_name);
        return;
    }

    interface_name = g_key_file_get_string(backend_file,
                                           TOP_SECTION,
                                           INTERFACE_NAME,
                                           NULL);

    if (interface_name == NULL) {
        g_warning("backend_file (%s) has interface_name == NULL in %s group.",
                  file_name, TOP_SECTION);
        g_strfreev(groups);
        g_free(module_name);
        g_free(node_name);
        return;
    }

    if (strchr(interface_name, '.') == NULL) {
        gchar *interface_name_short = interface_name;
        interface_name = g_strconcat(INTERFACE_PREFIX, interface_name, NULL);
        g_free(interface_name_short);
    }

    if (!g_dbus_is_interface_name(interface_name)) {
        g_warning("Incorrect interface name (%s) in backend file (%s).",
                interface_name, file_name);
        g_strfreev(groups);
        g_free(interface_name);
        g_free(node_name);
        g_free(module_name);
        return;
    }

    gchar *action_id = g_key_file_get_string(backend_file,
                                             TOP_SECTION,
                                             ACTION_ID,
                                             NULL);

    /* If action_id == NULL or it is not correct we use interface name. */
    if (action_id == NULL) {
        action_id = g_strdelimit (g_strdup (interface_name), "_", '-');
    } else if (!is_correct_action_id(action_id)) {
        g_warning("Incorrect actiont_id (%s) in backend file (%s).",
                   interface_name, file_name);
        g_free(action_id);
        action_id = g_strdelimit (g_strdup (interface_name), "_", '-');
    }

    /* If the node already contains the interface name we exit, because a node
       can't contains two interfaces with one name. */
    if (g_hash_table_contains(backends_data, node_name)) {
        interfaces =
                   (GHashTable *) g_hash_table_lookup(backends_data, node_name);

        if (g_hash_table_contains(interfaces, interface_name)) {
            g_warning("backends_data: %s already contains interface named %s.",
                      node_name, interface_name);
            g_strfreev(groups);
            g_free(interface_name);
            g_free(node_name);
            g_free(module_name);
            g_free(action_id);
            return;
        }
    } else {
        interfaces = g_hash_table_new_full(g_str_hash,
                                           g_str_equal,
                                           g_free,
                                           interface_info_free);

        g_hash_table_replace(backends_data, g_strdup(node_name), interfaces);
    }

    for (int i = 0; i < length; i++) {
        /* We skip group named Alterator Entry, we need only methods. */
        if (!g_strcmp0(groups[i], TOP_SECTION)) {
            continue;
        }

        g_debug("groups[%d] = %s", i, groups[i]);

        if (!is_correct_name(groups[i])) {
            g_warning("Invalid method name: %s.", groups[i]);
            continue;
        }

        /* An interface is created if it has at least one method. */
        if (!g_hash_table_contains(interfaces, interface_name)) {
            interface_info = g_new0(InterfaceObjectInfo, 1);
            if (interface_info == NULL) {
                g_warning("save_info_from_backend_file(), g_new0() for"
                           "interface_info returned NULL.");
                break;
            }

            /* interface_introspection and vtable are created and populated
               inside a specific module */
            interface_info->interface_introspection = NULL;
            interface_info->interface_vtable = NULL;
            interface_info->module_name = g_strdup(module_name);
            interface_info->action_id = g_strdup(action_id);
            interface_info->thread_limit = thread_limit;
            interface_info->methods = g_hash_table_new_full(g_str_hash,
                                                          g_str_equal,
                                                          g_free,
                                                          hashtable_free);

            g_hash_table_replace(interfaces,
                                 g_strdup(interface_name),
                                 interface_info);
        }

        GHashTable * method_data = g_hash_table_new_full(g_str_hash,
                                                         g_str_equal,
                                                         g_free,
                                                         g_free);

        gsize method_keys_number;
        gchar **method_keys = g_key_file_get_keys(backend_file,
                                                  groups[i],
                                                  &method_keys_number,
                                                  NULL);

        for (int ii = 0; ii < method_keys_number; ii++) {
            gchar *str = g_key_file_get_string(backend_file,
                                               groups[i],
                                               method_keys[ii],
                                               NULL);

            g_hash_table_replace(method_data, g_strdup(method_keys[ii]), str);
        }

        if (method_keys) {
            g_strfreev(method_keys);
        }

        g_hash_table_replace(interface_info->methods,
                             g_strdup(groups[i]),
                             method_data);
    }

    g_strfreev(groups);
    g_free(node_name);
    g_free(action_id);
    g_free(interface_name);
    g_free(module_name);
}

static gboolean read_data_from_backend_files(const gchar *key_files_path,
                                             GPtrArray   *backend_file_names)
{
    GKeyFile *backend_file = NULL;
    gchar* backend_file_path = NULL;
    gboolean is_loaded = FALSE;
    GError *error = NULL;
    gchar* file_name = NULL;

    if (backend_file_names == NULL) {
        return FALSE;
    }

    if (backends_data == NULL) {
        backends_data = g_hash_table_new_full(g_str_hash,
                                              g_str_equal,
                                              g_free,
                                              hashtable_free);
    }

    for (int i = 0; i < backend_file_names->len; i++) {
        file_name = (gchar *)g_ptr_array_index(backend_file_names, i);
        if (file_name == NULL) {
            continue;
        }

        backend_file = g_key_file_new();
        backend_file_path = g_strconcat(key_files_path, file_name, NULL);
        is_loaded = g_key_file_load_from_file(backend_file,
                                              backend_file_path,
                                              G_KEY_FILE_NONE,
                                              &error);

        if (error) {
            g_warning("Backend file load error(%s): %s.",
                      file_name, error->message);
            g_key_file_free(backend_file);
            g_free(backend_file_path);
            g_error_free(error);
            error = NULL;
            file_name = NULL;
            is_loaded = FALSE;
            continue;
        }

        if (is_loaded) {
            save_info_from_backend_file(backend_file, file_name);
        }

        g_key_file_free(backend_file);
        g_free(backend_file_path);
        error = NULL;
        file_name = NULL;
        is_loaded = FALSE;
    }

    if (backends_data != NULL) {
        return TRUE;
    } else {
        return FALSE;
    }
}

static gboolean alterator_manager_backends_load(const gchar *key_files_path) {
    gboolean result = FALSE;
    GPtrArray *backend_file_names = read_backend_file_names(key_files_path);
    if (backend_file_names == NULL || key_files_path == NULL) {
        return FALSE;
    }

    if (backend_file_names->len == 0) {
        g_debug(".backend files not found (%s).", key_files_path);
    }

    result = read_data_from_backend_files(key_files_path, backend_file_names);
    if (!result) {
        g_warning("read_data_from_backend_files() returned FALSE.");
    }

    g_ptr_array_free(backend_file_names, TRUE);

    return result;
}

gboolean alterator_manager_backends_init(gboolean is_session) {
    gboolean result = FALSE;
    gboolean res = FALSE;

    if (is_session) {
        res = alterator_manager_backends_load(KEY_FILES_PATH_USER);
        if (res) {
            result = res;
        } else {
            g_warning("alterator_manager_backends_load() returned FALSE "
                      "for %s.", KEY_FILES_PATH_USER);
        }

        res = alterator_manager_backends_load(KEY_FILES_PATH_ETC_USER);
        if (res) {
            result = res;
        } else {
            g_warning("alterator_manager_backends_load() returned FALSE "
                      "for %s.", KEY_FILES_PATH_ETC_USER);
        }
    } else {
        res = alterator_manager_backends_load(KEY_FILES_PATH_SYS);
        if (res) {
            result = res;
        } else {
            g_warning("alterator_manager_backends_load() returned FALSE "
                      "for %s.", KEY_FILES_PATH_SYS);
        }

        res = alterator_manager_backends_load(KEY_FILES_PATH_SYS_NEW);
        if (res) {
            result = res;
        } else {
            g_warning("alterator_manager_backends_load() returned FALSE "
                      "for %s.", KEY_FILES_PATH_SYS_NEW);
        }

        res = alterator_manager_backends_load(KEY_FILES_PATH_ETC_SYS);
        if (res) {
            result = res;
        } else {
            g_warning("alterator_manager_backends_load() returned FALSE "
                      "for %s.", KEY_FILES_PATH_ETC_SYS);
        }

        res = alterator_manager_backends_load(KEY_FILES_PATH_ETC_SYS_NEW);
        if (res) {
            result = res;
        } else {
            g_warning("alterator_manager_backends_load() returned FALSE "
                      "for %s.", KEY_FILES_PATH_ETC_SYS_NEW);
        }
    }

    return result;
}

GHashTable* alterator_manager_backends_get_data(void) {
    return backends_data;
}
